# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Management of job options
"""
import sys
import yaml

# sentinel value; used when running HiveJob as a script
_READ_ARGS_FROM_SYS_ARGV = '_READ_ARGS_FROM_SYS_ARGV'


def process_args(parser, args):
    """Process the comman-line or program arguments
    and check for a config file with extra options
    """
    # arguments can be provided on command line
    if args == _READ_ARGS_FROM_SYS_ARGV:
        opt_args = sys.argv[1:]
    else:
        opt_args = args

    # first pass at the arg list
    options, rem_args = parser.parse_args(opt_args)

    # if a config file was supplied, then modify args list and reprocess
    if options.config_path is not None:
        cf = YamlConfig(options.config_path)
        merged_args = cf.merge_config_file_args(opt_args, options.runner)
        options, rem_args = parser.parse_args(merged_args)

    return (options, rem_args)


class YamlConfig():
    """Use a yaml file to provide options to your jobs
    """

    # list of allowed config file options
    # FIXME not all these options are implemented
    OPTION_MAP = {
        'aws_access_key_id': '--aws-access-key-id',
        'aws_secret_access_key': '--aws-secret-access-key',

        'aws_availability_zone': '--aws-availability-zone',  # not implemented
        'aws_region': '--aws-region ',  # not implemented
        's3_endpoint': '--s3-endpoint',  # not implemented
        'emr_endpoint': '--emr-endpoint ',  # not implemented

        'hive_version': '--hive-version',
        'ami_version': '--ami-version',

        'ec2_master_instance_type': '--ec2-master-instance-type',
        'ec2_instance_type': '--ec2-instance-type',
        'num_ec2_instances': '--num-ec2-instances',

        's3_log_uri': '--s3-log-uri',
        's3_scratch_uri': '--s3-scratch-uri',
        's3_sync_wait_time': '--s3-sync-wait-time',
        'check_emr_status_every': '--check-emr-status-every',
        }

    def __init__(self, path):
        """Initialise the object with a file path
        """
        f = open(path, "r")
        self.conf = yaml.load(f)

    def merge_config_file_args(self, args, runner):
        """Merge options from the config file with supplied arguments.
        Options from command line or program take precedence over
        options provided in the config file
        """
        try:
            if runner == 'emr':
                opts = self.conf['runners']['emr']
            elif runner == 'local':
                opts = self.conf['runners']['local']
        except KeyError:
            return args  # no extra options available

        append_opts = []
        for k, v in self.OPTION_MAP.iteritems():
            if k in opts and v not in args:
                append_opts.append(v)
                append_opts.append(opts[k])
        return args + append_opts



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to manage set up and running of Hivejobs
"""
import sys
import logging

from optparse import Option
# from optparse import OptionError
# from optparse import OptionGroup
from optparse import OptionParser
from apiarist.emr import EMRRunner
from apiarist.local import LocalRunner
from apiarist.util import log_to_null
from apiarist.util import log_to_stream
from apiarist.conf import process_args

logger = logging.getLogger(__name__)


class ArgumentMissingError(Exception):
    pass


class HiveJobLauncher(object):

    OPTION_CLASS = Option

    def __init__(self, job_name, args=None):

        self.job_name = job_name

        # allow argument to be passed in to be used
        # to compose the script (variables/parameters)
        self._passthrough_options = []

        self.option_parser = OptionParser(usage=self._usage(),
                                          option_class=self.OPTION_CLASS,
                                          add_help_option=False)
        self.configure_options()

        # this class is instantiated without args
        # when run in standalone/command-line mode
        if args:

            # build options from args provided
            # this may load the yaml config also
            self.options, rem_args = process_args(self.option_parser, args)

            #  after named args have been removed by OptionParser
            #  only remaining argument is the location of input data
            try:
                self.input_data = rem_args[0]
            except IndexError:
                raise ArgumentMissingError("must provide path to source data")

        # Make it possible to redirect stdin, stdout, and stderr, for testing
        # TODO create sandbox()
        self.stdin = sys.stdin
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    def execute(self):
        self.run_job()

    @classmethod
    def set_up_logging(cls, quiet=False, verbose=False, stream=None):
        """Set up logging when running from the command line.
        This will also set up a null log handler for boto, so we don't get
        warnings if boto tries to log about throttling and whatnot.
        """
        if quiet:
            log_to_null(name='apiarist')
            log_to_null(name='__main__')
        else:
            log_to_stream(name='apiarist', debug=verbose, stream=stream)
            log_to_stream(name='__main__', debug=verbose, stream=stream)
        log_to_null(name='boto')

    def run_job(self):
        """Run the job
        """
        self.set_up_logging(quiet=self.options.quiet,
                            verbose=self.options.verbose,
                            stream=self.stderr)
        #  log the options being used
        logger.info("Launching job {0}".format(self.job_name))
        with self.make_runner() as runner:
            runner.run()

    def make_runner(self):
        """Make a runner based on arguments provided
        """
        if self.options.runner == 'emr':
            kwargs = self.emr_job_runner_kwargs()
            logger.info("Initating EMR runner: {}".format(kwargs))
            return EMRRunner(**kwargs)
        else:
            kwargs = self.local_job_runner_kwargs()
            logger.info("Initiating local runner: {}".format(kwargs))
            return LocalRunner(**kwargs)

    def local_job_runner_kwargs(self):
        return {
            'input_path': self.input_data,
            'output_dir': self.options.output_dir,
            'hive_query': self.hive_query(),
            'job_name': self.job_name,
            'temp_dir': self.options.scratch_dir,
            'no_output': self.options.no_output,
            }

    def emr_job_runner_kwargs(self):
        slave_instance_type = self.options.slave_instance_type
        master_instance_type = (self.options.master_instance_type or
                                slave_instance_type)
        return {
            'aws_access_key_id': self.options.aws_access_key_id,
            'aws_secret_access_key': self.options.aws_secret_access_key,
            'input_path': self.input_data,
            'output_dir': self.options.output_dir,
            'hive_query': self.hive_query(),
            'scratch_uri': self.options.scratch_uri,
            'log_path': self.options.log_uri,
            'job_name': self.job_name,
            'master_instance_type': master_instance_type,
            'slave_instance_type': slave_instance_type,
            'num_instances': self.options.num_instances,
            'ami_version': self.options.ami_version,
            'hive_version': self.options.hive_version,
            's3_sync_wait_time': self.options.s3_sync_wait_time,
            'check_emr_status_every': self.options.check_emr_status_every,
            'temp_dir': self.options.scratch_dir
            }

    def hive_query(self):
        # implemented in subclass HiveJob
        raise NotImplementedError

    def configure_options(self):
        """Define the arguments for this script
        """
        # the running mode - local or EMR
        self.option_parser.add_option(
            '-r', dest='runner', action='store', default='local'
            )

        # allow YAML config file
        self.option_parser.add_option(
            '--conf-path', dest='config_path', action='store', default=None
            )

        # where will the output files go?
        self.option_parser.add_option(
            '--output-dir', dest='output_dir', action='store', default=False
            )

        # AWS credentials
        # FIXME use these options when provided (overrides ENV or config file);
        self.option_parser.add_option(
            '--aws-access-key-id', dest='aws_access_key_id',
            action='store', default=None
        )
        self.option_parser.add_option(
            '--aws-secret-access-key', dest='aws_secret_access_key',
            action='store', default=None
        )

        # where to put the logs & temp files
        self.option_parser.add_option(
            '--s3-log-uri', dest='log_uri',
            action='store', default=False
            )
        self.option_parser.add_option(
            '--s3-scratch-uri', dest='scratch_uri',
            action='store', default=False
            )
        self.option_parser.add_option(
            '--local-scratch-dir', dest='scratch_dir',
            action='store', default=False
            )

        # hadoop config, job concurrency, and instance sizes
        self.option_parser.add_option(
            '--ec2-master-instance-type', dest='master_instance_type',
            action='store', default=False
            )
        self.option_parser.add_option(
            '--ec2-instance-type', dest='slave_instance_type',
            action='store', default='m3.xlarge'
            )
        self.option_parser.add_option(
            '--num-ec2-instances', dest='num_instances',
            action='store', default=2
            )
        self.option_parser.add_option(
            '--ami-version', dest='ami_version',
            action='store', default='latest'
            )
        self.option_parser.add_option(
            '--hive-version', dest='hive_version',
            action='store', default='latest'
            )

        # job runner options
        self.option_parser.add_option(
            '--s3-sync-wait-time', dest='s3_sync_wait_time',
            action='store', default=5
        )
        self.option_parser.add_option(
            '--check-emr-status-every', dest='check_emr_status_every',
            action='store', default=30
        )
        self.option_parser.add_option(
            '--no-output', dest='no_output',
            action='store_true', default=False
        )

        # logging options
        self.option_parser.add_option(
            '--quiet', dest='quiet',
            action='store_true', default=False
            )
        self.option_parser.add_option(
            '--verbose', dest='verbose',
            action='store_true', default=False
            )

    def add_passthrough_option(self, *args, **kwargs):
        """Add a section in the Job to specify options passed to
        the job to compile the query.

        Override the `configure_options` method in your job.

            def configure_options(self):
                super(MRYourJob, self).configure_options()
                self.add_passthrough_option('--start-date',
                                            default='2013-1-1',
                                            dest='startdate')

        Your job will now be able to access `self.options.startdate`
        """
        pass_opt = self.option_parser.add_option(*args, **kwargs)
        self._passthrough_options.append(pass_opt)

    @classmethod
    def _usage(cls):
        """Command line usage string for this class"""
        return ("usage: python [script path|executable path|--help]")



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from apiarist.s3 import upload_file_to_s3


class UnknownSerdeError(Exception):
    pass


class Serde(object):
    """Class to manage the Hive table serde jars
    """
    JARS_DIR = os.path.join(os.path.dirname(__file__), 'jars')
    CSV_JAR = 'csv-serde-1.1.2-0.11.0-all.jar'

    def __init__(self, serde='csv', s3_base_path=None):
        if serde == 'csv':
            self.type = 'CSV'
            self.jar = os.path.abspath(os.path.join(self.JARS_DIR,
                                                    self.CSV_JAR))
        else:
            raise UnknownSerdeError

        # base path for s3 files
        if s3_base_path is None:
            s3_base_path = os.environ.get('S3_SCRATCH_URI')
        self._s3_base_path = s3_base_path

    def s3_path(self):
        """get or create a location on S3 for the serde jar"""
        if 'CSV_SERDE_JAR_S3' in os.environ:
            # known location on S3
            serde = os.environ['CSV_SERDE_JAR_S3']
        else:
            if self._s3_base_path is None:
                raise ValueError("must specify the S3 scratch URI")
            # ensure the jar is up on S3
            jar_path = self._s3_base_path + 'jars/csv-serde.jar'
            upload_file_to_s3(self.jar, jar_path)
            os.environ['CSV_SERDE_JAR_S3'] = serde = jar_path
        return serde



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from apiarist.serde import Serde
from apiarist import InvalidHiveJobException
import os
import logging

logger = logging.getLogger(__name__)


def generate_hive_script_file(script, local_script_file):
    f = open(local_script_file, 'w')
    f.writelines(script)
    f.close()


def get_script_file_location(job_id, tmp_dir=None):
    if tmp_dir:
        td = tmp_dir
    elif 'APIARIST_TMP_DIR' in os.environ:
        td = os.environ['APIARIST_TMP_DIR']
    else:
        td = '/tmp/'
    return td + job_id + '.hql'


class HiveQuery(object):
    """Class to manage the contruction of the Hive query
    including the boilerplate to load data
    Can return a Hive query for local or EMR execution.
    """

    #  NOTE: There are no date data types; dates are treated as strings.
    #  There are several date functions which operate on strings.
    HIVE_TYPES = [
        'TINYINT',  # 1 byte
        'SMALLINT'  # 2 byte
        'INT'       # 4 byte
        'BIGINT'    # 8 byte
        'FLOAT'     # 4 byte (single precision floating point numbers)
        'DOUBLE'    # 8 byte (double precision floating point numbers)
        'BOOLEAN'   # you know what these are
        'STRING',   # up to 2GB
        ]

    def __init__(self, hive_job):
        """Initalise the query with a HiveJob object
        """
        logger.debug("initialising HiveQuery")
        try:
            logger.debug("setting input table name")
            self.table_name = hive_job.table()
            logger.debug("setting output table name")
            self.results_table_name = self.table_name + "_results"
            logger.debug("setting input columns")
            self.input_columns = hive_job.input_columns()
            logger.debug("setting output columns")
            self.output_columns = hive_job.output_columns()
            logger.debug("setting input control characters")
            self.input_control_chars = (hive_job.INFILE_DELIMITER_CHAR,
                                        hive_job.INFILE_QUOTE_CHAR,
                                        hive_job.INFILE_ESCAPE_CHAR
                                        )
            logger.debug("setting output control characters")
            self.output_control_chars = (hive_job.OUTFILE_DELIMITER_CHAR,
                                         hive_job.OUTFILE_QUOTE_CHAR,
                                         hive_job.OUTFILE_ESCAPE_CHAR
                                         )
            logger.debug("setting plain qeury content")
            self.query = hive_job.plain_query()

        except AttributeError, e:
            logger.error("Error encoutered setting query attributes")
            logger.error("Check the logic in your job")
            logger.error(e)
            ex = "HiveQuery expects a HiveJob-compatible oject"
            raise InvalidHiveJobException(ex)

        # ensure we have a semi-colon to terminate the select query
        if self.query[-1:] != ';':
            self.query += ';'

        # ensure the table name is included in the query
        if self.table_name not in self.query:
            raise ValueError("query does not contain a reference to the table")

        #  TODO validate the input/output columns for
        #  proper data types and reserved keywords

    def __repr__(self):
        return "HiveQuery:{}...".format(self.query[:80])

    def _csv_serde_jar(self, s3_scratch_uri):
        """Using a JAR for serialisation/deserialisation in the Hive tables
        """
        serde = Serde('csv', s3_scratch_uri)
        return serde.s3_path()

    def local_hive_script(self, data_source, output_dir, temp_table_dir):
        """generate a hive script to execute on the local hive server
        generates a CSV file via a hive textfile table
        """
        # boilerplate
        parts = [
            #  serde required before attempting drop tables
            "ADD JAR {0};".format(Serde('csv').jar),
            "DROP TABLE {0};".format(self.table_name),
            "DROP TABLE {0};".format(self.results_table_name),
            ]
        #  add the table in which we'll load the source data
        parts += self.create_table_ddl(self.table_name,
                                       self.input_columns,
                                       temp_table_dir,
                                       self.input_control_chars)
        #  add statement to load the source data into this table
        parts.append("LOAD DATA LOCAL INPATH '{0}' INTO TABLE {1};".format(
            data_source, self.table_name))
        #  add a table to select the results into (for CSV formatting)
        parts += self.create_table_ddl(self.results_table_name,
                                       self.output_columns,
                                       output_dir,
                                       self.output_control_chars)
        #  insert the results of the supplied query into this table
        parts.append("INSERT INTO TABLE {0}".format(
            self.results_table_name, self.query))
        #  and finally, the query
        parts.append(self.query)
        #  return a string that can be written to a file and run on Hive
        return "\n".join(parts)

    def emr_hive_script(self, data_source, output_dir, temp_table_dir,
                        s3_scratch_uri=None):
        """Generate the complete Hive script for EMR
        igenerates a set of comma-delimited files via a hive textfile table
        """
        # boilerplate
        parts = [
            "ADD JAR {0};".format(self._csv_serde_jar(s3_scratch_uri)),
            "SET hive.exec.compress.output=false;"
            ]
        # add the table in which we'll load the source data
        parts += self.create_table_ddl(self.table_name,
                                       self.input_columns,
                                       temp_table_dir,
                                       self.input_control_chars)
        # add statement to load the source data into this table
        parts.append("LOAD DATA INPATH '{0}' INTO TABLE {1};".format(
            data_source, self.table_name))
        # add a table to select the results into (for CSV formatting)
        parts += self.create_table_ddl(self.results_table_name,
                                       self.output_columns,
                                       output_dir,
                                       self.output_control_chars)
        # insert the results of the supplied query into this table
        parts.append("INSERT INTO TABLE {0}".format(self.results_table_name))
        # and finally, the query
        parts.append(self.query)
        # return a string that can be written to a file and run on Hive
        return "\n".join(parts)

    def create_table_ddl(self, name, columns, location, control_chars):
        """Create a Hive table to store CSV data
        """
        cols = self._column_ddl(columns)
        return [
            "CREATE EXTERNAL TABLE {0} ({1})".format(name, cols),
            "ROW FORMAT serde 'com.bizo.hive.serde.csv.CSVSerde'",
            "WITH serdeproperties (",
            r'"separatorChar" = "{}",'.format(control_chars[0]),
            r'"quoteChar" = "{}",'.format(control_chars[1]),
            r'"escapeChar" = "{}"'.format(control_chars[2]),
            ")",
            "STORED AS TEXTFILE",
            "LOCATION '{0}';".format(location)
            ]

    def _column_ddl(self, columns):
        """Get column defintions for Hive table
        """
        cols = ["`{0}` {1}".format(col[0], col[1]) for col in columns]
        return ", ".join(cols)



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to inherit your HiveJobs from. See README for more info
"""
import logging
import re

from apiarist.launch import HiveJobLauncher
from apiarist.script import HiveQuery
from apiarist.conf import _READ_ARGS_FROM_SYS_ARGV
from apiarist import InvalidHiveJobException

logger = logging.getLogger(__name__)


class HiveJob(HiveJobLauncher):

    # defaults for the input file
    INFILE_DELIMITER_CHAR = r','
    INFILE_QUOTE_CHAR = r'\"'
    INFILE_ESCAPE_CHAR = r'\\'

    # defaults for the output file
    OUTFILE_DELIMITER_CHAR = r','
    OUTFILE_QUOTE_CHAR = r'\"'
    OUTFILE_ESCAPE_CHAR = r"\\"

    def __init__(self, args=None):
        super(HiveJob, self).__init__(self._job_name(), args)

    def hive_query(self):
        """Get the Hive script object based on provided params
        """
        return HiveQuery(self)

    def plain_query(self):
        """Condense spaces"""
        try:
            q = self.query()
            return re.sub(r"\s+", " ", q).strip()
        except TypeError, e:
            logger.error("Could not get HiveQuery content")
            logger.debug(e)
            raise InvalidHiveJobException("Could not get HiveQuery content")

    def _job_name(self):
        return self.__class__.__name__

    @classmethod
    def run(cls):
        """Entry point for running job from the command-line.
        """
        hive_job = cls(args=_READ_ARGS_FROM_SYS_ARGV)
        hive_job.execute()

    #  methods which define the query logic and I/O

    def input_columns(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def output_columns(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def table(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def query(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" S3/storage utility methods """

import os
import re
import logging
from boto.s3.connection import S3Connection
from boto.s3.key import Key

logger = logging.getLogger(__name__)


class MissingDataException(Exception):
    pass


def get_conn(aws_access_key_id=None, aws_secret_access_key=None):
    k = aws_access_key_id or os.environ['AWS_ACCESS_KEY_ID']
    s = aws_secret_access_key or os.environ['AWS_SECRET_ACCESS_KEY']
    return S3Connection(k, s)


def copy_s3_file(source, destination,
                 aws_access_key_id=None, aws_secret_access_key=None):
    """ Copy an S3 object from one location to another
    """
    dest_bucket, dest_key = parse_s3_uri(destination)
    source_bucket, source_key = parse_s3_uri(source)
    conn = get_conn(aws_access_key_id, aws_secret_access_key)
    logger.info("Copying S3 source files. This may take some time.")
    if is_dir(source):
        s_bkt = conn.get_bucket(source_bucket)
        d_bkt = conn.get_bucket(dest_bucket)
        objs = get_bucket_list(s_bkt, source_key)
        if len(objs) == 0:
            raise MissingDataException("supplied path is empty")
        for i, k in enumerate(get_bucket_list(s_bkt, source_key)):
            new_key = dest_key + str(i)
            logger.debug("copying {0}/{1} to {2}/{3}".format(source_bucket,
                                                             k.key,
                                                             dest_bucket,
                                                             new_key))
            d_bkt.copy_key(new_key, source_bucket, k.key)
        return destination + '/'
    else:
        bkt = conn.get_bucket(dest_bucket)
        logger.debug("copying {0}/{1} to {2}/{3}".format(source_bucket,
                                                         source_key,
                                                         dest_bucket,
                                                         dest_key))
        return bkt.copy_key(dest_key, source_bucket, source_key)


def upload_file_to_s3(file_path, s3_path,
                      aws_access_key_id=None, aws_secret_access_key=None):
    """Create an S3 object from the contents of a local file
    """
    s3_bucket, s3_key = parse_s3_uri(s3_path)
    conn = get_conn(aws_access_key_id, aws_secret_access_key)
    bkt = conn.get_bucket(s3_bucket)
    k = Key(bkt)
    k.key = s3_key
    return k.set_contents_from_filename(file_path)


def parse_s3_uri(uri):
    """Parse an S3 uri from: s3://bucketname/some/other/path/info/
    to:
        bucket = bucketname
        key = some/other/path/info
    """
    m = re.search(r'(s3://)([A-Za-z0-9_-]+)/(\S*)', uri)
    if m:
        return (m.group(2), m.group(3))
    else:
        return None


def obj_type(key):
    """If it is a 'dir' it will end with a slash
    otherwise it is a 'file'
    """
    if key[-1:] == '/':
        return 'directory'
    else:
        return 'file'


def is_dir(key):
    return obj_type(key) == 'directory'


def concatenate_keys(source_dir, destination_key,
                     aws_access_key_id=None, aws_secret_access_key=None):
    """Concatenate all the files in a bucket
    using multipart upload feature of S3 API.
    NOTE: this only works when all files are above 5MB
    """
    s_bucket, s_key = parse_s3_uri(source_dir)
    d_bucket, d_key = parse_s3_uri(destination_key)
    conn = get_conn(aws_access_key_id, aws_secret_access_key)
    s_bk = conn.get_bucket(s_bucket)
    d_bk = conn.get_bucket(d_bucket)
    mp = d_bk.initiate_multipart_upload(d_key)
    for i, k in enumerate(get_bucket_list(s_bk, s_key)):
        mp.copy_part_from_key(s_bucket, k.key, i+1)
    mp.complete_upload()


def get_bucket_list(bucket, key):
    """ list items in a bucket that match given key """
    # ignore key if zero bytes
    return [k for k in bucket.list(key) if k.size > 0]



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to manage EMR config and job running
"""
import os
import hashlib
import re
import time
import datetime
import logging

import boto
from boto.emr.step import HiveStep
from boto.emr.step import InstallHiveStep
from boto.emr.connection import EmrConnection
from apiarist.s3 import copy_s3_file, is_dir, upload_file_to_s3
from apiarist.script import generate_hive_script_file, get_script_file_location

logger = logging.getLogger(__name__)


class EMRRunner():

    def __init__(self, job_name=None, input_path=None, hive_query=None,
                 output_dir=None, scratch_uri=None, log_path=None,
                 ami_version=None, hive_version=None, num_instances=None,
                 master_instance_type=None, slave_instance_type=None,
                 aws_access_key_id=None, aws_secret_access_key=None,
                 s3_sync_wait_time=5, check_emr_status_every=30,
                 temp_dir=None):

        self.job_name = job_name
        self.job_id = self._generate_job_id()
        self.start_time = time.time()

        # AWS credentials can come from arguments or environment
        # set AWS credentials if supplied (override ENV)
        if aws_access_key_id:
            os.environ['AWS_ACCESS_KEY_ID'] = aws_access_key_id
            self.aws_access_key_id = aws_access_key_id
        else:
            logger.debug("Getting AWS access key from ENV")
            self.aws_access_key_id = os.environ['AWS_ACCESS_KEY_ID']

        if aws_secret_access_key:
            os.environ['AWS_SECRET_ACCESS_KEY'] = aws_secret_access_key
            self.aws_secret_access_key = aws_secret_access_key
        else:
            logger.debug("Getting AWS secret key from ENV")
            self.aws_secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']

        self.s3_sync_wait_time = s3_sync_wait_time
        self.check_emr_status_every = check_emr_status_every

        # I/O for job data
        self.input_path = input_path
        self.output_dir = output_dir

        # is the input multiple files in a 'directory'?
        try:
            self.input_is_dir = is_dir(input_path)
        except TypeError:
            self.input_is_dir = False

        # the Hive script object
        self.hive_query = hive_query

        #  EMR options
        self.master_instance_type = master_instance_type
        self.slave_instance_type = slave_instance_type
        self.ami_version = ami_version
        self.hive_version = hive_version
        self.num_instances = num_instances

        # S3 'scratch' directory
        if scratch_uri:
            self.base_path = scratch_uri
            os.environ['S3_SCRATCH_URI'] = scratch_uri
        else:
            self.base_path = os.environ['S3_SCRATCH_URI']

        # allow alternate logging path
        self.log_path = log_path or self.base_path + 'logs/'
        # other temp files live in a jobID bucket
        self.job_files = self.base_path + self.job_id + '/'
        self.data_path = self.job_files + 'data'
        if self.input_is_dir:
            self.data_path += '/'
        self.table_path = self.job_files + 'tables/'
        self.script_path = self.job_files + 'script.hql'
        self.output_path = self.output_dir or self.job_files + 'output/'

        # a local temp dir is used to write the script
        self.local_script_file = get_script_file_location(self.job_id,
                                                          temp_dir)

        logger.info("JobID {0}, started at {1}".format(self.job_id,
                                                       self.start_time))

    def _generate_hive_script(self, data_source):
        """Write the HQL to a local (temp) file
        """
        hq = self.hive_query.emr_hive_script(data_source, self.output_path,
                                             self.table_path)
        generate_hive_script_file(hq, self.local_script_file)

    def _generate_job_id(self):
        """Create a unique job run identifier
        """
        run_id = self.job_name + str(time.time())
        digest = hashlib.md5(run_id).hexdigest()
        return 'hj-' + digest

    def _generate_and_upload_hive_script(self):
        self._generate_hive_script(self.data_path)
        upload_file_to_s3(self.local_script_file, self.script_path)

    #  hooks for the with statement ###

    def __enter__(self):
        """Don't do anything special at start of with block"""
        return self

    def __exit__(self, type, value, traceback):
        """Call self.cleanup() at end of with block."""
        self.cleanup()

    def run(self):
        """Run the Hive job on EMR cluster
        """
        #  copy the data source to a new object
        #  (Hive deletes/moves the original)
        copy_s3_file(self.input_path, self.data_path)

        # and create the hive script
        self._generate_and_upload_hive_script()

        logger.info("Waiting {} seconds for S3 eventual consistency".format(
                    self.s3_sync_wait_time))
        time.sleep(self.s3_sync_wait_time)

        # TODO more options like setting aws region
        conn = EmrConnection(self.aws_access_key_id,
                             self.aws_secret_access_key)

        setup_step = InstallHiveStep(self.hive_version)
        run_step = HiveStep(self.job_name, self.script_path)

        jobid = conn.run_jobflow(
            self.job_name,
            self.log_path,
            action_on_failure='CANCEL_AND_WAIT',
            master_instance_type=self.master_instance_type,
            slave_instance_type=self.slave_instance_type,
            ami_version=self.ami_version,
            num_instances=self.num_instances)

        conn.add_jobflow_steps(jobid, [setup_step, run_step])

        self._wait_for_job_to_complete(conn, jobid)

        logger.info("Output file is in: {0}".format(self.output_path))

    def cleanup(self):
        # TODO _ remove scratch dirs?
        logger.info("cleaning up ... ")

    # wait for job and log status (from mrjob)
    # this method extracted from mrjob.job
    def _wait_for_job_to_complete(self, conn, jobid):
        """Wait for the job to complete, and raise an exception if
        the job failed.

        Also grab log URI from the job status (since we may not know it)
        """
        success = False
        chk_status_freq = self.check_emr_status_every
        # opts = {'check_emr_status_every': 30}
        # s3_logs = self.log_path
        emr_job_start = self.start_time

        while True:
            # don't antagonize EMR's throttling
            logger.debug('Waiting {0} seconds'.format(chk_status_freq))
            time.sleep(chk_status_freq)

            job_flow = conn.describe_jobflow(jobid)

            job_state = job_flow.state
            reason = getattr(job_flow, 'laststatechangereason', '')

            # find all steps belonging to us, and get their state
            step_states = []
            running_step_name = ''
            total_step_time = 0.0
            step_nums = []  # step numbers belonging to us. 1-indexed
            # lg_step_num_mapping = {}

            steps = job_flow.steps or []
            # latest_lg_step_num = 0
            for i, step in enumerate(steps):
                # if LOG_GENERATING_STEP_NAME_RE.match(
                # posixpath.basename(getattr(step, 'jar', ''))):
                #    latest_lg_step_num += 1

                # ignore steps belonging to other jobs
                if not step.name.startswith(self.job_name):
                    continue

                step_nums.append(i + 1)
                # if LOG_GENERATING_STEP_NAME_RE.match(
                # posixpath.basename(getattr(step, 'jar', ''))):
                #    lg_step_num_mapping[i + 1] = latest_lg_step_num

                step.state = step.state
                step_states.append(step.state)
                if step.state == 'RUNNING':
                    running_step_name = step.name

                if hasattr(step, 'startdatetime') and \
                   hasattr(step, 'enddatetime'):

                    start_time = iso8601_to_timestamp(step.startdatetime)
                    end_time = iso8601_to_timestamp(step.enddatetime)
                    total_step_time += end_time - start_time

            if not step_states:
                raise AssertionError("Can't find our steps in the job flow!")

            # if all our steps have completed, we're done!
            if all(state == 'COMPLETED' for state in step_states):
                success = True
                break

            # if any step fails, give up
            if any(state in ('FAILED', 'CANCELLED') for state in step_states):
                break

            # (the other step states are PENDING and RUNNING)

            # keep track of how long we've been waiting
            running_time = time.time() - emr_job_start

            # otherwise, we can print a status message
            if running_step_name:
                logger.info("Job launched {0} ago, status {1}: {2} ({3})".
                            format(int(running_time), job_state, reason,
                                   running_step_name))

                # if self._show_tracker_progress:
                #    try:
                #        tracker_handle = urllib2.urlopen(self._tracker_url)
                #        tracker_page = ''.join(tracker_handle.readlines())
                #        tracker_handle.close()
                #        # first two formatted percentages, map then reduce
                #        map_complete, reduce_complete = [
                #            float(complete) for complete
                #            in JOB_TRACKER_RE.findall(tracker_page)[:2]]
                #        logger.info(' map %3d%% reduce %3d%%' % (
                #                 map_complete, reduce_complete))
                #    except:
                #       logger.info('Unable to load progress from job tracker')
                #        # turn off progress for rest of job
                #        self._show_tracker_progress = False
                # once a step is running, it's safe to set up the ssh tunnel to
                # the job tracker
                # job_host = getattr(job_flow, 'masterpublicdnsname', None)
                # if job_host and opts['ssh_tunnel_to_job_tracker']:
                #    self.setup_ssh_tunnel_to_job_tracker(job_host)

            # other states include STARTING and SHUTTING_DOWN
            elif reason:
                logger.info("Job launched {0} ago, status {1}: {2}".format(
                    int(running_time), job_state, reason))
            else:
                logger.info("Job launched {0} ago, status {1}".format(
                    int(running_time), job_state))

        if success:
            logger.info('Job completed.')
            logger.info("Running time was {0}".format(total_step_time))
            logger.info("(excludes time spent waiting for the EC2 instances)")
        else:
            msg = 'Job on job flow {0} failed with status {1}: {2}'.format(
                  job_flow.jobflowid, job_state, reason)
            logger.info(msg)

            cause = False
            # TODO resurrect this code to recover reason for failure
            # if self._s3_job_log_uri:
            #    logger.info('Logs are in %s' % self._s3_job_log_uri)
            # look for a Python traceback
            # cause = self._find_probable_cause_of_failure(

            if cause:
                # log cause, and put it in exception
                cause_msg = []  # lines to log and put in exception
                cause_msg.append('Probable cause of failure (from {0}):'.format
                                 (cause['log_file_uri']))
                cause_msg.extend(line.strip('\n') for line in cause['lines'])
                if cause['input_uri']:
                    cause_msg.append('(while reading from {0})'.format(
                                     cause['input_uri']))
                for line in cause_msg:
                    logger.info(line)

                # add cause_msg to exception message
                msg += '\n' + '\n'.join(cause_msg) + '\n'

            raise Exception(msg)

#  AWS Date-time parsing

# sometimes AWS gives us seconds as a decimal, which we can't parse
# with boto.utils.ISO8601
SUBSECOND_RE = re.compile('\.[0-9]+')

# Thu, 29 Mar 2012 04:55:44 GMT
RFC1123 = '%a, %d %b %Y %H:%M:%S %Z'


def iso8601_to_timestamp(iso8601_time):
    iso8601_time = SUBSECOND_RE.sub('', iso8601_time)
    try:
        return time.mktime(time.strptime(iso8601_time, boto.utils.ISO8601))
    except ValueError:
        return time.mktime(time.strptime(iso8601_time, RFC1123))


def iso8601_to_datetime(iso8601_time):
    iso8601_time = SUBSECOND_RE.sub('', iso8601_time)
    try:
        return datetime.strptime(iso8601_time, boto.utils.ISO8601)
    except ValueError:
        return datetime.strptime(iso8601_time, RFC1123)



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to manage local job execution
"""
import os
import subprocess
import hashlib
import time
import shutil
import logging
from apiarist.script import generate_hive_script_file, get_script_file_location

logger = logging.getLogger(__name__)


class LocalRunner():
    """Handles running the Hive script on
    a local Hive installation.
    """

    def __init__(self, job_name=None, input_path=None,
                 hive_query=None, output_dir=None,
                 temp_dir=None, no_output=False):

        #  TODO test for Hive installation

        self.job_name = job_name
        self.job_id = self._generate_job_id()
        self.start_time = time.time()

        # I/O for job data
        self.scratch_dir = self.get_local_scratch_dir(temp_dir)
        self.stream_output = (not no_output)

        self.data_path = self.scratch_dir + 'data'
        self.table_path = self.scratch_dir + 'table'
        self.input_path = os.path.abspath(input_path)
        if output_dir:
            self.output_dir = os.path.abspath(output_dir) + '/' + self.job_id
        else:
            self.output_dir = self.scratch_dir + 'output'

        # the Hive script object
        self.hive_query = hive_query
        self.local_script_file = get_script_file_location(self.job_id,
                                                          self.scratch_dir)

    def get_local_scratch_dir(self, temp_dir=None):
        if temp_dir:
            tmp_path = temp_dir + self.job_id + '/'
        elif 'APIARIST_TMP_DIR' in os.environ:
            tmp_path = os.environ['APIARIST_TMP_DIR'] + self.job_id + '/'
        else:
            tmp_path = "{0}/.apiarist/{1}/".format(os.environ['HOME'],
                                                   self.job_id)
        return tmp_path

    def _ensure_local_scratch_dir_exists(self):
        if not os.path.exists(self.scratch_dir):
            os.makedirs(self.scratch_dir)

    def run(self):
        """Run the hive query against a local hive installation (*nix only)
        """
        # prepare files
        self._ensure_local_scratch_dir_exists()
        self._copy_input_data()
        self._generate_hive_script()
        # execute against hive server
        cmd = ["hive -f {}".format(self.local_script_file)]
        logger.info("running HIVE script with: {}".format(cmd))
        hql = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        stdout = hql.communicate()
        if stdout[1] is not None:
            logger.info(stdout)
        # observe and report
        self._wait_for_job_to_complete()

    def _copy_input_data(self):
        shutil.copyfile(self.input_path, self.data_path)

    def _wait_for_job_to_complete(self):
        # TODO - wait until there are files in this dir
        cmd = ["cat {}/*".format(self.output_dir)]
        cat = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT, shell=True)
        stdout, stderr = cat.communicate()
        if self.stream_output:
            logger.info("\nQuery output ------->\n")
            print(stdout)  # query results to STDOUT

    def _generate_hive_script(self):
        """Write the HQL to a local (temp) file
        """
        hq = self.hive_query.local_hive_script(self.data_path,
                                               self.output_dir,
                                               self.table_path)
        generate_hive_script_file(hq, self.local_script_file)

    def _generate_job_id(self):
        """Create a unique job run identifier
        """
        run_id = self.job_name + str(time.time())
        digest = hashlib.md5(run_id).hexdigest()
        return 'hj-' + digest

    #  hooks for the with statement ###

    def __enter__(self):
        """Don't do anything special at start of with block"""
        s = self
        return s

    def __exit__(self, type, value, traceback):
        """Call self.cleanup() at end of with block."""
        self.cleanup()

    def cleanup(self):
        # TODO _ remove scratch dirs?
        pass



# Copyright 2014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions that have no external dependencies."""

import sys
import logging


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def log_to_null(name=None):
    """Set up a null handler for the given stream, to suppress
    no handlers could be found" warnings."""
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())


def log_to_stream(name=None, stream=None, format=None, level=None,
                  debug=False):
    """Set up logging."""
    if level is None:
        level = logging.DEBUG if debug else logging.INFO

    if format is None:
        format = '%(message)s'

    if stream is None:
        stream = sys.stderr

    handler = logging.StreamHandler(stream)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter(format))

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)
